#ifndef AIBOEDGEDETECTOR_H_DEFINED
#define AIBOEDGEDETECTOR_H_DEFINED

/**
   Classe per il calcolo del Sobel Edge Detection sui robot Aibo della
   Sony. Utilizza alcuni campi non documentati presenti nelle immagini
   in formato RAW; stando ad alcuni documenti non ufficiali le
   immagini in formato RAW hanno la seguente forma:

   <table cellspacing=0 valign="center">
   <tr>
   <th rowspan=5>height</th>
   <th>width</th>
   <th>width</th>
   <th>width</th>
   <th>width</th>
   <th>width</th>
   <th>width</th>
   </tr>
   
   <tr>
   <td>YYYY</td>
   <td>CrCrCrCr</td>
   <td>CbCbCbCb</td>
   <td>GradXGradXGradX</td>
   <td>GradYGradYGradY</td>
   <td>MagMagMag</td>
   </tr>
   
   <tr>
   <td>YYYY</td><td>CrCrCrCr</td><td>CbCbCbCb</td><td>GradXGradXGradX</td><td>GradYGradYGradY</td><td>MagMagMag</td>
   </tr>
   
   <tr>
   <td colspan=6><center>...</center></td>
   </tr>
   
   <tr>
   <td>YYYY</td><td>CrCrCrCr</td><td>CbCbCbCb</td><td>GradXGradXGradX</td><td>GradYGradYGradY</td><td>MagMagMag</td>
   </tr> 
   </table> 
   
   <p> Da alcuni test risulta che il campo Mag e' privo di
   significato, quindi viene ricalcolato sulla base dei due gradienti
   con l'operazione: abs(MAG)=abs(GraX)+abs(GradY).<br>
   </p>
   
   <p>
   Questa classe risulta molto efficiente nel calcolo dell'Edge
   Detection, basandosi su valori ritornati dall'hardware degli Aibo,
   anche se l'immagine risultante risulta essere leggermente inferiore
   in qualita' rispetto a quella ottenuta applicando l'algoritmo con
   routine software.
   </p>
 */
class AiboEdgeDetector{
 public:
  /**
     Costruttore di default. Costruisce un oggetto non utilizzabile.
   */
  AiboEdgeDetector() :
    mIAmValid(false),
    mProcessedImage(0) {};
  
  /**
     Costruttore. Costruisce un oggetto per il calcolo dell'Edge
     Detection data altezza e larghezza dell'immagine.

     @param width largezza dell'immagine in pixel

     @param height altezza dell'immagine in pixel.
   */
  AiboEdgeDetector(int width, int height);
  
  /**
     Distruttore di default.
   */
  ~AiboEdgeDetector();

  /**
     Calcola la Magnitude del Sobel Edge Detection sull'immagine
     passata come argomento. La matrice ritornata assueme valori da
     0(assenza di bordi) a 255(massimo bordo) (dai test e' difficile
     avere bordi con magniute superiore a 100).L'argomento dev'essere
     un'immagine Aibo in formato RAW, come ritornata dal metodo
     OFbkImageVectorData::getData().

     @param pImage immagine RAW

     @return array const contenente il magnitude dell'immagine di
     dimensione widthxheight byte.
   */
  const unsigned char* getMagnitude(unsigned char* pImage);
 
  /**
     Calcola Gradiente X del Sobel Edge Detection sull'immagine
     passata come argomento. La matrice ritornata assume valori
     [-128,127]. L'argomento dev'essere un'immagine Aibo in formato
     RAW, come ritornata dal metodo OFbkImageVectorData::getData().
     
     @param pImage immagine RAW

     @return array const contenente il magnitude dell'immagine di
     dimensione widthxheight byte.
   */
  const char* getGradientX(unsigned char* pImage);

  /**
     Calcola Gradiente Y del Sobel Edge Detection sull'immagine
     passata come argomento. La matrice ritornata assume valori
     [-128,127]. L'argomento dev'essere un'immagine Aibo in formato
     RAW, come ritornata dal metodo OFbkImageVectorData::getData().
     
     @param pImage immagine RAW
     
     @return array const contenente il magnitude dell'immagine di
     dimensione widthxheight byte.
  */
  const char* getGradientY(unsigned char* pImage);
  

 private:
  /**
     Flag per valutare se l'oggetto e' valido e inizializzato o meno.
   */
  bool mIAmValid;
  
  /**
     Larghezza dell'immagine in pixel.
   */
  int mWidth;  
  /**
     Altezza dell'immagine in pixel.
   */
  int mHeight; 
  /**
     Dimensione dell'immagine in byte.
   */
  int mImageSize;
  	       
  /**
     Array contenente l'immagine processata di dimensione mImageSize.
   */
  unsigned char* mProcessedImage;

  /**
     Dimesione di una riga dell'immagine in formato raw.
   */
  int mSomeWidth;
};

#endif
